//
// Static screensaver
// --------------------
// (c) 1996 Lucian Wischik
//
// There are a couple of issues with a DirectDraw screensaver
// 1. You cannot do SetMode in response to WM_CREATE. DirectDraw
//    simply doesn't support it. This screensaver does setmode
//    in SCRM_TIMER, for the very first timer message it recieves.
// 2. When the mode is changed, the main window typically recieves
//    WM_ACTIVATE(FALSE) and WM_NCACTIVATE(FALSE). Under normal
//    circumstances, this would terminate the screensaver. So,
//    this one intercepts those messages if the flag 'ischangingmode'
//    indicates that it is appropriate.
// 3. When we want a verify-password dialog or something, we must
//    switch out of ModeX. This is because ModeX does not support
//    Windows GDI. This is done by responding to SCRM_DIALOG(TRUE)
//    for 'suspend', and SCRM_DIALOG(FALSE) for 'resume'. This
//    screensaver keeps 'suspendbits' to keep an internal copy
//    of the state of the screensaver, and 'backdropbits' for the
//    thing that actually get displayed during this time.
// 4. This screensaver uses ChangePalette() to colour-cycle.
// 5. For a decent Plus! screensaver, you need to be able to do
//    the preview. You could do that with a smaller chunk of DirectDraw.
//    This screen saver chooses to use the flexibility of the TBits
//    library to use DIBSections for previews and ModeX for full-screen:
//    and because TBits is so wonderful, no code changes are needed!
//

#define WIN32_LEAN_AND_MEAN
#define STRICT
#include <windows.h>
#include <stdlib.h>
#include <mmsystem.h>
#include <ddraw.h>
#include "../../lib/scrplus.h"
#include "bits.h"
#pragma hdrstop


HINSTANCE hInstance;


class TStaticData {
public:
  TBits bits;           // The main pair of surfaces for animation.
  TBits suspendbits;    // Where we keep a record while we have a dialog popped up
  TBits backdropbits;   // The backdrop for while a dialog is popped up
  int frame;            // How many cycles have elapsed so far
  BOOL wantddx;         // wantddx, cx and cy are calculated in Start()
  int cx,cy;            // for the desired parameters of the image
  BOOL ischangingmode;  // Set to TRUE when changing, to avoid spurious WM_ACTIVATE
  BOOL isstarted;       // because we can only init DirectDraw after WM_CREATE
  BOOL issuspended;     // Whether the thing has been frozen because of a dialog

  TStaticData();
  ~TStaticData();
  void Start(HWND);
  void Stop(HWND);
  unsigned int Step(HWND hwnd);
  void DoPalette();
  void Suspend(HWND);
  void Resume(HWND);
};


TStaticData::TStaticData()
{ frame=0; wantddx=FALSE; cx=0; cy=0; ischangingmode=FALSE; isstarted=FALSE;
  issuspended=FALSE; }
TStaticData::~TStaticData() {bits.Reset();}

void TStaticData::Start(HWND hwnd)
{ BOOL isfull=FALSE; if (GetParent(hwnd)==NULL) isfull=TRUE;
  if (isfull)
  { cx=320; cy=200;
    wantddx=TRUE;
  }
  else
  { RECT rc; GetClientRect(hwnd,&rc);
    cx=rc.right-rc.left; cy=rc.bottom-rc.top;
    wantddx=FALSE;
  }
  ischangingmode=TRUE;
  bits.EnableSubclass(TRUE,NULL);
  bits.DibPainter.PositionAtPoint(0,0);
  HRESULT err=bits.Create(hwnd,wantddx,2,cx,cy,8,TRUE);
  ischangingmode=FALSE;
  if (err!=BITS_OK) return;
  isstarted=TRUE;
}

void TStaticData::Stop(HWND) { bits.Reset(); }

void TStaticData::Suspend(HWND hwnd)
{ if (issuspended) return;
  if (!bits.isddx) return;
  issuspended=TRUE;
  suspendbits.Reset();
  suspendbits.CreateDibCopy(hwnd,bits.surface[1]);
  ischangingmode=TRUE;
  bits.Reset();
  UINT oldval;
  SystemParametersInfo(SPI_SCREENSAVERRUNNING,1,&oldval,0);
  // because leaving DirectDraw would have suddenly enabled
  // ctrl-alt-delete and task switching. We want to disable it again.
  int scx=GetSystemMetrics(SM_CXSCREEN);
  int scy=GetSystemMetrics(SM_CYSCREEN);
  POINT tl; tl.x=0; tl.y=0; ClientToScreen(hwnd,&tl);
  POINT br; br.x=scx; br.y=scy; ClientToScreen(hwnd,&br);
  MoveWindow(hwnd,tl.x,tl.y,br.x,br.y,FALSE);
  backdropbits.Reset();
  backdropbits.EnableFullSubclass(FALSE);
  backdropbits.CreateDibCopy(hwnd,suspendbits.surface[0]);
  backdropbits.Flip();
  ischangingmode=FALSE;
}

void TStaticData::Resume(HWND hwnd)
{ if (!issuspended) return;
  bits.Reset();
  bits.CopyPaletteFrom(suspendbits);
  ischangingmode=TRUE;
  backdropbits.Reset();
  HRESULT err=bits.Create(hwnd,wantddx,2,cx,cy,8,TRUE);
  ischangingmode=FALSE;
  if (err!=BITS_OK) return;
  bits.surface[1].CopyBitsFrom(suspendbits.surface[0]);
  suspendbits.Reset();
  bits.Flip();
  issuspended=FALSE;
}

unsigned int TStaticData::Step(HWND)
{ if (issuspended) return 100;
  frame++;
  TBitsSurface *bs=bits.LockSurface(1);
  if (bs!=NULL)
  { LPDWORD addr=bs->data.lpdw;
    int width=bits.cx/4;
    for (int y=0; y<bits.cy; y++)
    { DWORD myrand=rand(); myrand=NEXTRAND2(myrand);
      for (int x=0; x<width; x++)
      { myrand=NEXTRAND(myrand); *addr++=myrand;
      }
    }
  }
  bits.UnlockSurface(bs);
  DoPalette();
  bits.Flip();
  return 1;
}

void TStaticData::DoPalette()
{ BOOL goingdown;
  if ((frame%2000)<1000) goingdown=TRUE; else goingdown=FALSE;
  if (goingdown)
  { int nz=50; if (nz>30) { nz=0; for (int i=0; i<50; i++)
    	 { int j=(rand()%255)+1;
    		 if (bits.LogPal.Entries[j].peBlue<2) {bits.LogPal.Entries[j].peBlue=0;nz++;} else bits.LogPal.Entries[j].peBlue=(char)(bits.LogPal.Entries[j].peBlue-2); } }
    if (nz>30) { nz=0; for (int i=0; i<50; i++) { int j=(rand()%255)+1;
    		 if (bits.LogPal.Entries[j].peGreen<2) {bits.LogPal.Entries[j].peGreen=0;nz++;} else bits.LogPal.Entries[j].peGreen=(char)(bits.LogPal.Entries[j].peGreen-2); } }
    if (nz>30) { for (int i=0; i<50; i++) { int j=(rand()%255)+1;
    		 if (bits.LogPal.Entries[j].peRed<2) {bits.LogPal.Entries[j].peRed=0; } else bits.LogPal.Entries[j].peRed=(char)(bits.LogPal.Entries[j].peRed-2); } }
  }
  else
  { int nz=50; if (nz>30) { nz=0; for (int i=0; i<50; i++) { int j=(rand()%255)+1;
    		 if (bits.LogPal.Entries[j].peBlue>253) {bits.LogPal.Entries[j].peBlue=255;nz++;} else bits.LogPal.Entries[j].peBlue=(char)(bits.LogPal.Entries[j].peBlue+2); } }
    if (nz>30) { nz=0; for (int i=0; i<50; i++) { int j=(rand()%255)+1;
     		 if (bits.LogPal.Entries[j].peGreen>253) {bits.LogPal.Entries[j].peGreen=255;nz++;} else bits.LogPal.Entries[j].peGreen=(char)(bits.LogPal.Entries[j].peGreen+2); } }
    if (nz>30) { for (int i=0; i<50; i++) { int j=(rand()%255)+1;
    		 if (bits.LogPal.Entries[j].peRed>253) {bits.LogPal.Entries[j].peRed=255;} else bits.LogPal.Entries[j].peRed=(char)(bits.LogPal.Entries[j].peRed+2); } }
  }
  bits.ChangePalette();
}

LRESULT CALLBACK MyScreenSaverProc(HWND hwnd,UINT msg,WPARAM wParam,LPARAM lParam)
{ TStaticData *sd=(TStaticData *)GetWindowLong(hwnd,GWL_USERDATA);
  if (sd==NULL && msg!=WM_CREATE) return DefScreenSaverProc(hwnd,msg,wParam,lParam);
  switch (msg) {
    case (WM_CREATE):
		{ TStaticData *sd=new TStaticData();
        SetWindowLong(hwnd,GWL_USERDATA,(LONG)sd);
      } break;
    case (WM_ACTIVATE):  case (WM_NCACTIVATE): case (WM_ACTIVATEAPP):
      { if (sd->ischangingmode) return DefWindowProc(hwnd,msg,wParam,lParam);
      } break;
    case (SCRM_TIMER):
      { if (!sd->isstarted) sd->Start(hwnd); return sd->Step(hwnd);
		}
    case (SCRM_DIALOG):
      { if ((BOOL)wParam) sd->Suspend(hwnd); else sd->Resume(hwnd);
      } break;
    case (WM_DESTROY):
      { sd->Stop(hwnd);
        SetWindowLong(hwnd,GWL_USERDATA,0);
        delete sd;
      } break;
  }
  return DefScreenSaverProc(hwnd,msg,wParam,lParam);
}



int WINAPI WinMain(HINSTANCE h,HINSTANCE,LPSTR,int)
{ hInstance=h;
  BitsInit(hInstance);
  TScrPrefs *p=ScrCreatePrefs();
  p->AllowActiveConfig=TRUE;
  p->AllowMute=FALSE;
  p->IsTrace=FALSE;
  p->IsDebug=FALSE;
#ifdef STATICCODE
  strcpy(p->StaticCode,STATICCODE);
#endif
#ifdef DYNAMICCODE
  strcpy(p->DynamicCode,DYNAMICCODE);
#endif
  ScrSetPrefs(p);

  int res=ScrExecute(hInstance,MyScreenSaverProc);

  BitsClose();
  return res;
}



